// Version 1.0 of Feb. 1, 2003
// All rights reserved.
// For conditions of use see the readme-file in this 
// package. Current version available at 
// www.lrz.de/~ekkehart --> downloads

// Program originally written in HP-Basic by
// Ekkehart Schlicht and ported to C by Robert Schlicht
// Version 1.0 of Feb. 1, 2003.

// To change convergency or the type of relaxation, look at the very end
// of the program.

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

// Die Funktion 'vector' erzeugt einen Vektor bzw. ndert seine Lnge.
// 'x' zeigt auf die Adresse des Vektors, die auf 0 gesetzt ist, wenn der
// Vektor die Lnge 0 hat, und die auf 0 gesetzt werden muss, bevor die
// Funktion das erste Mal fr einen Vektor aufgerufen wird. 'length' ist
// die neue Lnge des Vektors, die mitgespeichert wird. Wenn nicht
// gengend Speicherplatz frei ist, wird das Programm abgebrochen.
void vector(double ** x, unsigned int length)
{
    unsigned int *ptr;
    if (!x) abort();    // (ungltiges Argument)
    ptr = !*x? 0 : (unsigned int*)(void*)*x - 1;
    if (!length) {if (ptr) free(ptr); *x = 0; return;}
    if (ptr && *ptr == length) return;
    ptr = realloc(ptr, sizeof *ptr + length * sizeof **x);
    if (!ptr) abort();  // (nicht gengend Speicherplatz frei)
    *x = (double*)(void*)(ptr + 1);
    *ptr = length;
}

// Dieselbe Funktion fr eine Matrix (= Vektor aus Vektoren):
void matrix(double *** A, unsigned int length, unsigned int width)
{
    unsigned int *ptr, n;
    if (!A) abort();    // (ungltiges Argument)
    ptr = !*A? 0 : (unsigned int*)(void*)*A - 1;
    if (ptr) for (n=length; n<*ptr; n++) vector(&(*A)[n], 0);
    if (!length) {if (ptr) free(ptr); *A = 0; return;}
    n = !ptr? 0:*ptr;
    ptr = realloc(ptr, sizeof *ptr + length * sizeof **A);
    if (!ptr) abort();  // (nicht gengend Speicherplatz frei)
    *A = (double**)(void*)(ptr + 1);
    for (; n<length; n++) (*A)[n] = 0;
    *ptr = length;
    for (n=0; n<length; n++) vector(&(*A)[n], width);
}

// Das Makro 'LENGTH' ermittelt die Lnge eines Vektors oder einer Matrix:
#define LENGTH(x) ( !(x)? 0 : *((unsigned int*)(void*)(x) - 1) )




#ifndef WINDOWS

// Hauptprogramm (Konsolenversion):
int main(int argc, char *argv[])
{
    void sol(double **x, double *y, double **A, double *r, double *z);
    double var(double **x, double *y, double **A, double *r, double *s0, double *s, double *z);

    FILE *file;
    double **x=0, *y=0, **A=0, *r=0, *z=0, *s=0, s0;
    double v;
    unsigned int l, n, i, t;
    char c;
    
    // Datei ffnen
    if (argc < 2) return 0;	// (Dateiname nicht angegeben)
    file = fopen(argv[1], "rt");
    if (!file) return 1;    // (Fehler beim ffnen der Datei)

    // Daten einlesen (Matrix x und Vektor y erzeugen)
    n = 0;
    while (fscanf(file, "%*[^,\n]%c ", &c), c == ',') ++n;  // in erster Zeile Spalten zhlen
    --n;                                                    // n = Breite von x (erste Spalte wird ignoriert, zweite Spalte ist y)
    l = 0;                                                  // l = Lnge von x bzw. y
    while (!feof(file)) {                                       
        vector(&y, l+1); matrix(&x, l+1, n);                // Lnge von x und y vergrern
        fscanf(file, "%*[^,],%lf ", &y[l]);                 // erste Spalte berspringen und y-Wert einlesen
        for (i=0; i<n; i++) fscanf(file, ",%lf ", &x[l][i]);// Zeile von x einlesen
        l++;                                                // l um 1 erhhen
    }

    // Datei schlieen
    fclose(file);
    
    // Vektoren r, z, s und Matrix A erzeugen
    // (Der "Varianzvektor (s(0), s(1), ..., s(N))" ist hier s0, s[0], ..., s[n-1].)
    vector(&r, n), vector(&z, n), vector(&s, n), matrix(&A, l, n);
    
    // Anfangswerte fr s0, s und r festlegen
    s0 = 1;
    for (i=0; i<n; i++) s[i] = 1;
    for (i=0; i<n; i++) r[i] = s[i] / s0;


    // Normalgleichungssystem fr A lsen und iterierte
    // Varianzen und Varianzverhltnisse bestimmen
    i = 0;
    do {
        sol(x, y, A, r, z);
        v = var(x, y, A, r, &s0, s, z);
        if (!(++i%10)) printf("Iterationen: %u; Varianzquotient: %.6lf  \r", i, v);
    } while (v >= .001);
    printf("Iterationen: %u; Varianzquotient: %.6lf  \r", i, v);

            // Bemerkung: Die Konvergenzgeschwindigkeit kann in bestimmten
            // Situationen verbessert werden, indem man in der Funktion var(...)
            // eine zustzliche Zeile einfgt (s.u.).

    // Ergebnis anzeigen
    printf("\n\nErgebnismatrix A:\n\n");
    for (t=0; t<l; t++) {
        printf("      |");
        for (i=0; i<n; i++) printf("\t% .4lf", A[t][i]);
        printf("  |\n");
    }
    printf("\n\n");
    
    // Programm beenden, wenn kein zustzlicher Dateiname angegeben wurde
    if (argc < 3) return 0;

    // Sonst Ergebnis in Datei speichern und Programm beenden
    file = fopen(argv[2], "wt");
    if (!file) return 1;        // (Fehler beim Erzeugen der Datei)
    fprintf(file, "t,y");
    for (i=0; i<n; i++) fprintf(file, ",x(.;%u)", i+1);
    for (i=0; i<n; i++) fprintf(file, ",A(.;%u)", i+1);
    fprintf(file, "\n");
    for (t=0; t<l; t++) {
        fprintf(file, "%u,%lf", t+1, y[t]);
        for (i=0; i<n; i++) fprintf(file, ",%lf", x[t][i]);
        for (i=0; i<n; i++) fprintf(file, ",%lf", A[t][i]);
        fprintf(file, "\n");
    }

    return 0;
}

#else

#include <windows.h>	

// Prozedur fr Statusfenster (Windows-Version)
LRESULT CALLBACK WindowProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static TCHAR text[1024];

	switch (msg) {
		case 0xC000:
			lstrcpyn(text, (LPCTSTR)lParam, 1024);
			InvalidateRect(hwnd, 0, 0);
			break;

		case WM_PAINT: {
			PAINTSTRUCT ps; RECT rect; HDC hdc; HBITMAP hbmp; HFONT hfont;
			BeginPaint(hwnd, &ps);
			GetClientRect(hwnd, &rect);
			hbmp = SelectObject(hdc = CreateCompatibleDC(ps.hdc),
				CreateCompatibleBitmap(ps.hdc, rect.right, rect.bottom));
			FillRect(hdc, &rect, GetSysColorBrush(COLOR_BTNFACE));
			SetTextColor(hdc, GetSysColor(COLOR_BTNTEXT));
			SetBkMode(hdc, TRANSPARENT);
			hfont = SelectObject(hdc, GetStockObject(DEFAULT_GUI_FONT));
			DrawText(hdc, text, -1, &rect, DT_CENTER);
			SelectObject(hdc, hfont);
			BitBlt(ps.hdc, 0, 0, rect.right, rect.bottom, hdc, 0, 0, SRCCOPY);
			DeleteObject(SelectObject(hdc, hbmp)); DeleteDC(hdc);
			EndPaint(hwnd, &ps);
			break;
		}

		case WM_CLOSE: PostMessage(hwnd, 0xC000, 0, 0); break;

		default: return DefWindowProc(hwnd, msg, wParam, lParam);
	}
	return 0;
}

// Hauptprogramm (Windows-Version):
int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nCmdShow)
{
    void sol(double **x, double *y, double **A, double *r, double *z);
    double var(double **x, double *y, double **A, double *r, double *s0, double *s, double *z);

    FILE *file;
    double **x=0, *y=0, **A=0, *r=0, *z=0, *s=0, s0;
    double v;
    unsigned int l, n, i, t;
    char c;
	TCHAR text[1024]; OPENFILENAME ofn; WNDCLASS wc; HWND hwnd; MSG msg;

    // Datei ffnen
	ZeroMemory(&ofn, sizeof ofn); ofn.lStructSize = sizeof ofn;
	ofn.lpstrDefExt = TEXT("csv"); ofn.lpstrFilter = TEXT("CSV (*.csv)\0*.csv\0");
	ofn.nMaxFile = 1024; ofn.lpstrFile = text; *text = 0;
	ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_NOTESTFILECREATE | OFN_HIDEREADONLY;
	if (!GetOpenFileName(&ofn)) return 0;
	file = fopen(text, "rt");
    if (!file) return 1;    // (Fehler beim ffnen der Datei)

    // Daten einlesen (Matrix x und Vektor y erzeugen)
    n = 0;
    while (fscanf(file, "%*[^,\n]%c ", &c), c == ',') ++n;  // in erster Zeile Spalten zhlen
    --n;                                                    // n = Breite von x (erste Spalte wird ignoriert, zweite Spalte ist y)
    l = 0;                                                  // l = Lnge von x bzw. y
    while (!feof(file)) {                                       
        vector(&y, l+1); matrix(&x, l+1, n);                // Lnge von x und y vergrern
        fscanf(file, "%*[^,],%lf ", &y[l]);                 // erste Spalte berspringen und y-Wert einlesen
        for (i=0; i<n; i++) fscanf(file, ",%lf ", &x[l][i]);// Zeile von x einlesen
        l++;                                                // l um 1 erhhen
    }

    // Datei schlieen
    fclose(file);
    
    // Vektoren r, z, s und Matrix A erzeugen
    // (Der "Varianzvektor (s(0), s(1), ..., s(N))" ist hier s0, s[0], ..., s[n-1].)
    vector(&r, n), vector(&z, n), vector(&s, n), matrix(&A, l, n);
    
    // Anfangswerte fr s0, s und r festlegen
    s0 = 1;
    for (i=0; i<n; i++) s[i] = 1;
    for (i=0; i<n; i++) r[i] = s[i] / s0;

	// Statusfenster erzeugen
	ZeroMemory(&wc, sizeof wc); wc.lpfnWndProc = &WindowProc;
	wc.hInstance = GetModuleHandle(0); wc.hIcon = LoadIcon(0, IDI_ASTERISK);
	wc.hCursor = LoadCursor(0, IDC_ARROW); wc.lpszClassName = TEXT("Status");
	hwnd = CreateWindow(MAKEINTRESOURCE(RegisterClass(&wc)), TEXT("Status"),
		WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_VISIBLE,
		CW_USEDEFAULT, SW_SHOWDEFAULT, 200, 100, 0, 0, wc.hInstance, 0);

    // Normalgleichungssystem fr A lsen und iterierte
    // Varianzen und Varianzverhltnisse bestimmen
    i = 0;
    do {
        sol(x, y, A, r, z);
        v = var(x, y, A, r, &s0, s, z);
        if (!(++i%10)) {
        	sprintf(text, "\nIterationen: %u\n\nVarianzquotient: %.6lf", i, v);
        	SendMessage(hwnd, 0xC000, 0, (LPARAM)text);
			while (PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
				if (msg.message == 0xC000) goto destroy;
				else TranslateMessage(&msg), DispatchMessage(&msg);
        }
    } while (v >= .00001);

            // Bemerkung: Die Konvergenzgeschwindigkeit kann in bestimmten
            // Situationen verbessert werden, indem man in der Funktion var(...)
            // eine zustzliche Zeile einfgt (s.u.).

	// Statusfenster entfernen
    destroy: DestroyWindow(hwnd);


    // Ergebnis in Datei speichern und Programm beenden
	*text = 0;
	ofn.Flags = OFN_OVERWRITEPROMPT | OFN_PATHMUSTEXIST | OFN_NOREADONLYRETURN | OFN_HIDEREADONLY;
	if (!GetSaveFileName(&ofn)) return 0;
	file = fopen(text, "wt");
    if (!file) return 1;        // (Fehler beim Erzeugen der Datei)
    fprintf(file, "t,y");
    for (i=0; i<n; i++) fprintf(file, ",x(.;%u)", i+1);
    for (i=0; i<n; i++) fprintf(file, ",A(.;%u)", i+1);
    fprintf(file, "\n");
    for (t=0; t<l; t++) {
        fprintf(file, "%u,%lf", t+1, y[t]);
        for (i=0; i<n; i++) fprintf(file, ",%lf", x[t][i]);
        for (i=0; i<n; i++) fprintf(file, ",%lf", A[t][i]);
        fprintf(file, "\n");
    }

    return 0;
}

#endif


// Programm zur Lsung der Normalgleichungen.
// x, y und r werden vorgegeben, A wird bestimmt.
void sol(double **x, double *y, double **A, double *r, double *z)
{
    void cholband(double **A);
    void trace(double **C, double *z);
    void bandsol(double **A, double *x);

    static double **C, *a;  // (a und C werden nur einmal beim ersten Funktionsaufruf erzeugt.)
    unsigned int l, n, t, i, j, q;
    double p;
    
    // Feldgrenzen von A sind l und n;
    l = LENGTH(x);
    n = LENGTH(*x);

    if (!C) matrix(&C, l*n, n+1);
    
    for (j=1; j<l*n; j++) for (i=0; i<n+1; i++) C[j][i] = 0;

    for (t=0; t<l; t++) {
        p = y[t];
        q = t * n;
        for (i=0; i<n; i++) {
            A[t][i] = x[t][i] * p;
            for (j=0; j<=i; j++) C[q+i][n-i+j] = x[t][i] * x[t][j];
        }
    }
    
    for (i=0; i<n; i++) r[i] = 1/r[i];
    
    for (t=1; t<l; t++) {
        q = t * n;
        for (i=0; i<n; i++) C[q+i][0] = -r[i], C[q+i][n] += 2*r[i];
    }
    
    for (i=0; i<n; i++) C[i][n] += r[i], C[q+i][n] -= r[i], r[i] = 1/r[i];
    
    cholband(C);
    trace(C, z);

    // Matrix A in Vektor a umdimensionieren
    if (!a) vector(&a, l*n);
    {double *g; g = a; for (t=0; t<l; t++) for (i=0; i<n; i++) *g++ = A[t][i];}

    bandsol(C, a);
    
    // Umdimensionieren rckgngig machen
    {double *g; g = a; for (t=0; t<l; t++) for (i=0; i<n; i++) A[t][i] = *g++;}
}

// A wird mit der Transponierten von A (kurz gespeichert)
// berschrieben
void bandtrn(double **A)
{
    unsigned int m, n, p, i, j;

    n = LENGTH(A);
    m = LENGTH(*A);
    
    for (i=0; i<n; i++) for (j=0; j<m && j<n-i; j++) A[i][j] = A[i+j][m-1-j];
    for (j=1; j<m; j++) for (i=n-j; i<n; i++) A[i][j]=0;
}

// BANDSOL1 lst A*y=x und berschreibt x mit y:
// A ist untere Dreiecksmatrix kurz gespeichert
void bandsol1(double **A, double *x)
{
    unsigned int m, n, p, i, j;

    n = LENGTH(A);
    m = LENGTH(*A);
    p = m-1;
    
    for (i=0; i<n; i++) {
        for (j = i<p? p-i:0; j<p; j++) x[i] -= A[i][j] * x[j-(p-i)];
        x[i] /= A[i][p];
    }
}

// Wie BANDSOL1, nur fr obere Dreiecksmatrix
void bandsol2(double **A, double *x)
{
    unsigned int m, n, i, j;

    n = LENGTH(A);
    m = LENGTH(*A);
    
    for (i=n; i>0;) {i--;
        for (j=1; j<m && j<n-i; j++) x[i] -= A[i][j] * x[i+j];
        x[i] /= A[i][0];
    }
}

// bandsol lst (A*A')*y=x und berschreibt x mit y.
// A ist untere ist untere Dreiecksmatrix, kurz gespeichert.
void bandsol(double **A, double *x)
{
    bandsol1(A, x);
    bandtrn(A);
    bandsol2(A, x);
}

// Cholesky-Zerlegung der (kurz gespeicherten) Bandmatrix A
// A wird mit unterer Dreiecksmatrix berschrieben.
void cholband(double **A)
{
    unsigned int n, m, p, i, j, k;
    
    n = LENGTH(A);
    m = LENGTH(*A);
    p = m-1;

    for (i=0; i<n; i++) {
        for (j = i>p? i-p:0; j<i; j++) {
            for (k = i>p? i-p:0; k<j; k++) A[i][p+j-i] -= A[i][p+k-i]*A[j][p+k-j];
            A[i][p+j-i] /= A[j][p];
        }
        for (k = i>p? i-p:0; k<i; k++) A[i][p] -= A[i][p+k-i] * A[i][p+k-i];
        A[i][p] = sqrt(A[i][p]);
    }
}

// trace bestimmt z(i) = tr{P(i)*INV(CC')*P'(i)}
// fr i = 1 bis n, C untere Dreiecksmatrix, kurz gespeichert.
void trace(double **C, double *z)
{
    unsigned int m, n, l, i, j, t;
    static double *x, *y; // (Diese Vektoren werden nur einmal beim ersten Funktionsaufruf erzeugt.)
    double q;
    
    m = LENGTH(C);
    n = LENGTH(z);
    l = m / n;
    
    if (!x) vector(&x, m);
    if (!y) vector(&y, m);
    
    for (i=0; i<n; i++) {
        q = 0;
        for (j=0; j<m; j++) y[j] = 0;
        y[i] = 1;
        bandsol1(C, y);
        for (t=1; t<l; t++) {
            for (j=0; j<m; j++) x[j] = y[j], y[j] = 0;
            y[t*n+i] = 1;
            bandsol1(C, y);
            for (j=0; j<m; j++) x[j] = y[j] - x[j], q += x[j] * x[j];
        }
        z[i] = q;
    }

}

// var bestimmt neue Varianzverhltnisse und Varianzen
double var(double **x, double *y, double **A, double *r, double *s0, double *s, double *z)
{
    static double *u, *r0; // (Diese Vektoren werden nur einmal beim ersten Funktionsaufruf erzeugt.)
    unsigned int l, n, t, i;
    double p, q, f;
    
    l = LENGTH(x);
    n = LENGTH(*x);

    if (!u) vector(&u, l);
    if (!r0) vector(&r0, n);

    for (i=0; i<n; i++) r0[i] = r[i];

    for (t=0; t<l; t++) {
        u[t] = -y[t];
        for (i=0; i<n; i++) u[t] += A[t][i]*x[t][i];
    }

    q = 0; for (t=0; t<l; t++) q += u[t] * u[t];

    for (i=0; i<n; i++) {
        s[i] = 0;
        for (t=1; t<l; t++) f = A[t][i] - A[t-1][i], s[i] += f * f;
        q += s[i] / r[i];
    }
    *s0 = q / (l-n);

    p = 0;
    for (i=0; i<n; i++) {
        r[i] = (s[i] * (l-n) / q + z[i]) / (l-1);
        f = r[i] / r0[i];
//      r[i] = r0[i] * pow(f, 20);  // Mit dieser Zeile kann oft die Konvergenzgeschwindigkeit verbessert werden.
        s[i] = r[i] * *s0;
        f -= 1; if (f > p) p = f; else if (-f > p) p = -f;
    }

    return p;
}
